
import datetime
from vital_ai_vitalsigns.model.VITAL_Node import VITAL_Node


class DatabaseConnection(VITAL_Node):
        appID: str
        configString: str
        endpointType: str
        endpointURL: str
        organizationID: str
        password: str
        username: str
        readOnly: bool




from collections.abc import MutableSequence
from docarray import BaseDoc, DocList
from docarray.typing import NdArray
from vital_ai_vitalsigns.collection.rdf_collection_impl import RdfCollectionImpl
from vital_ai_vitalsigns.collection.vector_collection_impl import VectorCollectionImpl
from vital_ai_vitalsigns.model.GraphObject import GraphObject
from vital_ai_vitalsigns.model.VITAL_Edge import VITAL_Edge
from vital_ai_vitalsigns.query.result_element import ResultElement
from vital_ai_vitalsigns.query.result_list import ResultList
from vital_ai_vitalsigns.vitalsigns import VitalSigns
from typing import Optional


class GraphObjectDoc(BaseDoc):
    id: Optional[str]  # optional because of query docs
    URI: str = ''
    ClassURI: str = ''
    name: str = ''
    text: str = ''  # use to hold text in embedding vector
    embedding: NdArray[384]  # must match embedding size


class GraphCollection(MutableSequence):

    def __init__(self, data=None, use_rdfstore=True, use_vectordb=True, embedding_model_id='paraphrase-MiniLM-L3-v2'):
        self._use_rdfstore = use_rdfstore
        self._use_vectordb = use_vectordb
        self._embedding_model_id = embedding_model_id
        if data is None:
            self._data = []
        else:
            self._data = [item for item in data if isinstance(item, GraphObject)]
        self._uri_map = {}
        self._vector_properties = {}
        if use_vectordb is True:
            self._vectordb = VectorCollectionImpl(GraphObjectDoc)
        if use_rdfstore is True:
            self._rdfstore = RdfCollectionImpl()
        for item in self._data:
            if hasattr(item, 'URI'):

                self._uri_map[item.URI] = item

                if self._use_rdfstore is True:
                    obj_nt = item.to_rdf()
                    self._rdfstore.add_triples(obj_nt)

                if self._use_vectordb is True:
                    embedding_model = VitalSigns.get_embedding_model(self._embedding_model_id)

                    uri = str(item.URI)
                    name = str(item.name)

                    if len(self._vector_properties) == 0:
                        text = name
                    else:
                        text = self.get_object_text(item)

                    encoding = embedding_model.vectorize(text)

                    index_docs = [GraphObjectDoc(id=uri, URI=uri, name=name, text=text, embedding=encoding)]

                    self._vectordb.index_documents(DocList[GraphObjectDoc](index_docs))

    def __len__(self):
        return len(self._data)

    def __getitem__(self, index):
        return self._data[index]

    def __setitem__(self, index, value):
        if not isinstance(value, GraphObject):
            raise ValueError("All items must be instances of GraphObject or its subclasses")
        if hasattr(self._data[index], 'URI'):
            self._uri_map.pop(self._data[index].URI, None)
        if hasattr(value, 'URI'):
            self._uri_map[value.URI] = value
        self._data[index] = value

    def __delitem__(self, index):
        if hasattr(self._data[index], 'URI'):
            self._uri_map.pop(self._data[index].URI, None)
        del self._data[index]

    def insert(self, index, value):
        if not isinstance(value, GraphObject):
            raise ValueError("All items must be instances of GraphObject or its subclasses")
        self.pop(value.URI)
        if hasattr(value, 'URI'):
            self._uri_map[value.URI] = value
        self._data.insert(index, value)

    def get(self, uri, default=None):
        """Return the first GraphObject in the collection with a matching URI."""
        for item in self._data:
            if item.URI == uri:
                return item
        return default  # Or raise an exception if preferred

    def pop(self, uri, default=None):
        """Remove and return the first GraphObject in the collection with a matching URI.
        If no item is found, return 'default' if provided, or None."""
        for i, item in enumerate(self._data):
            if item.URI == uri:
                if self._use_rdfstore is True:
                    self._rdfstore.delete_triples(uri)
                if self._use_vectordb is True:
                    self._vectordb.remove_doc(uri)
                return self._data.pop(i)  # Removes and returns the item
        if default is not None:
            return default
        # raise ValueError("No item found with the specified URI.")
        return None

    def remove(self, uri, default=None):
        return self.pop(uri, default)

    def add(self, obj):
        if not isinstance(obj, GraphObject):
            raise ValueError("Item must be instances of GraphObject or its subclasses")
        self.pop(obj.URI)
        self._data.append(obj)
        if self._use_vectordb is True:

            vs = VitalSigns()

            embedding_model = vs.get_embedding_model(self._embedding_model_id)

            uri = str(obj.URI)
            name = str(obj.name)

            if len(self._vector_properties) == 0:
                text = name
            else:
                text = self.get_object_text(obj)

            encoding = embedding_model.vectorize(text)

            index_docs = [GraphObjectDoc(id=uri, URI=uri, name=name, text=text, embedding=encoding)]

            self._vectordb.index_documents(DocList[GraphObjectDoc](index_docs))

        if self._use_rdfstore is True:
            obj_nt = obj.to_rdf()
            self._rdfstore.add_triples(obj_nt)

    def add_objects(self, objects):
        """Add multiple GraphObjects to the collection after ensuring all objects are valid."""
        # First, verify all objects are instances of GraphObject or its subclasses
        if not all(isinstance(obj, GraphObject) for obj in objects):
            raise ValueError("All items must be instances of GraphObject or its subclasses")

        # If the check passes, add all objects to the collection
        for obj in objects:
            self.pop(obj.URI)
            self._data.append(obj)
            if self._use_vectordb is True:

                vs = VitalSigns()

                embedding_model = vs.get_embedding_model(self._embedding_model_id)

                uri = str(obj.URI)
                name = str(obj.name)

                if len(self._vector_properties) == 0:
                    text = name
                else:
                    text = self.get_object_text(obj)

                encoding = embedding_model.vectorize(text)

                index_docs = [GraphObjectDoc(id=uri, URI=uri, name=name, text=text, embedding=encoding)]

                self._vectordb.index_documents(DocList[GraphObjectDoc](index_docs))
            if self._use_rdfstore is True:
                obj_nt = obj.to_rdf()
                self._rdfstore.add_triples(obj_nt)

    def get_object_text(self, obj):
        text = ""
        class_uri = obj.get_class_uri()
        for obj_class_uri in self._vector_properties:
            if obj_class_uri == class_uri:
                prop_uri_list = self._vector_properties[obj_class_uri]
                for puri in prop_uri_list:
                    value = obj.get_property_value(puri)
                    text = text + " " + str(value)
                    # print(f"Embedding Text {class_uri} in {puri} {str(value)}")

        return text.strip()

    def remove_objects(self, uris):
        """Remove objects from the collection by a list of URI values."""
        to_remove_indexes = []
        for uri in uris:
            for i, item in enumerate(self._data):
                if item.URI == uri:
                    to_remove_indexes.append(i)
                    if self._use_rdfstore is True:
                        self._rdfstore.delete_triples(uri)
                    if self._use_vectordb is True:
                        self._vectordb.remove_doc(uri)
                    break  # Assuming URIs are unique, stop after finding the first match

        # Iterate in reverse order to avoid altering the indexes of items to be removed
        for i in sorted(to_remove_indexes, reverse=True):
            del self._data[i]

    def set_vector_properties(self, class_uri, property_uris):
        """Set the vector properties for a given class URI."""
        if not isinstance(property_uris, list):
            raise ValueError("property_uris must be a list")
        self._vector_properties[class_uri] = property_uris

    def get_vector_properties(self, class_uri):
        """Get the vector properties for a given class URI."""
        return self._vector_properties.get(class_uri, [])

    def remove_vector_properties(self, class_uri):
        """Remove the vector properties for a given class URI."""
        if class_uri in self._vector_properties:
            del self._vector_properties[class_uri]

    def get_vector_class_uris(self):
        return list(self._vector_properties.keys())

    def get_edges_incoming(self, uri):
        incoming_edges = [item for item in self._data if isinstance(item, VITAL_Edge) and item.edgeDestination == uri]
        return incoming_edges

    def get_edges_outgoing(self, uri):
        outgoing_edges = [item for item in self._data if isinstance(item, VITAL_Edge) and item.edgeSource == uri]
        return outgoing_edges

    def get_nodes_incoming(self, uri):
        """
        Return the node objects that have an outgoing edge to the given URI.
        """
        # Use a set to avoid duplicates and collect URIs of source nodes
        incoming_node_uris = {item.edgeSource for item in self._data if
                              isinstance(item, VITAL_Edge) and item.edgeDestination == uri}
        # Use the existing get method to convert URIs to node objects
        incoming_nodes = [self.get(node_uri) for node_uri in incoming_node_uris]
        return [node for node in incoming_nodes if node is not None]  # Filter out any None results

    def get_nodes_outgoing(self, uri):
        """
        Return the node objects that are the destination of an outgoing edge from the given URI.
        """
        outgoing_node_uris = {item.edgeDestination for item in self._data if
                              isinstance(item, VITAL_Edge) and item.edgeSource == uri}
        outgoing_nodes = [self.get(node_uri) for node_uri in outgoing_node_uris]
        return [node for node in outgoing_nodes if node is not None]  # Filter out any None results

    def search(self, query: str, limit: int = 10):

        vs = VitalSigns()

        embedding_model = vs.get_embedding_model(self._embedding_model_id)

        query_embedding = embedding_model.vectorize(query)

        results = self._vectordb.search(query_embedding, limit)

        if results is None:
            return None

        if len(results.matches) == 0:
            return []

        result_list = ResultList()

        count = 0
        for r in results.matches:
            score = results.scores[count]
            uri = r.URI
            obj = self.get(uri)
            re = ResultElement(obj, score)
            result_list.append(re)
            count = count + 1

        return result_list

    def sparql_query(self, sparql_query: str):
        return self._rdfstore.query_graph(sparql_query)

    # TODO add to_json() to return json list

# allow metadata for class for a list of properties to use
# in text when generating vector, with default being "name" property


# to add:
# hypernode and hyperedge cases




import datetime
from vital_ai_vitalsigns.model.VITAL_Node import VITAL_Node


class VitalServiceConfig(VITAL_Node):
        appID: str
        configString: str
        connectionError: str
        connectionState: str
        defaultSegmentName: str
        key: str
        organizationID: str
        targetAppID: str
        targetOrganizationID: str
        uriGenerationStrategy: str
        primary: bool





import datetime
from vital_ai_vitalsigns_core.model.VitalServiceConfig import VitalServiceConfig


class VitalServiceIndexedDBConfig(VitalServiceConfig):
        graphQueries: str
        selectQueries: str




from .IProperty import IProperty


class OtherProperty(IProperty):
    def __init__(self, value: str):
        super().__init__(value)

    def __bool__(self):
        return bool(self.value)

    def __getattr__(self, attr):
        return getattr(self.value, attr)

    def __str__(self):
        return str(self.value)

    def __rshift__(self, other):
        return self == other

    def to_json(self):
        return {"value": self.value}




import datetime
from vital_ai_vitalsigns_core.model.VitalServiceConfig import VitalServiceConfig


class VitalServiceSqlConfig(VitalServiceConfig):
        dbType: str
        endpointURL: str
        password: str
        poolInitialSize: int
        poolMaxTotal: int
        username: str





import datetime
from vital_ai_vitalsigns.model.VITAL_Node import VITAL_Node


class DomainModel(VITAL_Node):
        appID: str
        backwardCompVersion: str
        defaultPackageValue: str
        domainOWL: str
        domainOWLHash: str
        organizationID: str
        preferredImportVersions: str
        versionInfo: str
        preferred: bool




from .IProperty import IProperty


class IntegerProperty(IProperty):
    def __init__(self, value: int):
        super().__init__(value)

    def __bool__(self):
        return bool(self.value)

    def __getattr__(self, attr):
        return getattr(self.value, attr)

    def __int__(self):
        return int(self.value)

    def __eq__(self, other):
        if isinstance(other, int):
            return self.value == other
        elif isinstance(other, IntegerProperty):
            return self.value == other.value
        return NotImplemented

    def __rshift__(self, other):
        return self == other

    def to_json(self):
        return {"value": self.value}



from datetime import datetime
import rdflib


class IProperty:
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return str(self.value)

    def to_json(self):
        return {"value": self.value}

    def to_rdf(self):
        if isinstance(self.value, datetime):
            value = self.value.isoformat()
            datatype = rdflib.XSD.dateTime
        elif isinstance(self.value, int):
            value = str(self.value)
            datatype = rdflib.XSD.integer
        elif isinstance(self.value, float):
            value = str(self.value)
            datatype = rdflib.XSD.float
        elif isinstance(self.value, bool):
            value = str(self.value)
            datatype = rdflib.XSD.boolean
        else:
            value = str(self.value)
            datatype = rdflib.XSD.string
        return {"value": value, "datatype": datatype}




class ClassProperty:

    def __init__(self, class_uri: str, property_uri: str):
        self._class_uri = class_uri
        self._property_uri = property_uri

    # use for cases like: VITAL_Node.name
    # which can be used for query constraints




from abc import ABC, abstractmethod
import json
import rdflib
from rdflib import Graph, Literal, URIRef
from vital_ai_vitalsigns.impl.vitalsigns_impl import VitalSignsImpl
from vital_ai_vitalsigns.model.trait.PropertyTrait import PropertyTrait
from vital_ai_vitalsigns.model.properties.BooleanProperty import BooleanProperty
from vital_ai_vitalsigns.model.properties.LongProperty import LongProperty
from vital_ai_vitalsigns.model.properties.StringProperty import StringProperty
from vital_ai_vitalsigns.model.properties.URIProperty import URIProperty


class AttributeComparisonProxy:
    def __init__(self, cls, name):
        self.cls = cls
        self.name = name

    def __eq__(self, value):
        print(f"Comparing {self.cls.__name__}.{self.name} with {value}")
        # Add your custom logic here
        return False  # Placeholder for the example


class GraphObjectMeta(type):
    def __setattr__(self, name, value):
        print(f"Setting class attribute {name} to {value}")
        super().__setattr__(name, value)

    def __getattr__(self, name):
        return AttributeComparisonProxy(self, name)


class GraphObject(metaclass=GraphObjectMeta):
    _allowed_properties = []

    @classmethod
    def get_allowed_properties(cls):
        return GraphObject._allowed_properties

    def __init__(self):
        super().__setattr__('_properties', {})

    def __setattr__(self, name, value) -> bool:
        # print(name)
        if name == 'URI':
            if value is None:
                self._properties.pop('http://vital.ai/ontology/vital-core#URIProp', None)
            else:
                self._properties['http://vital.ai/ontology/vital-core#URIProp'] = VitalSignsImpl.create_property_with_trait(URIProperty, 'http://vital.ai/ontology/vital-core#URIProp', value)
            return
        for prop_info in self.get_allowed_properties():
            uri = prop_info['uri']
            prop_class = prop_info['prop_class']
            trait_class = VitalSignsImpl.get_trait_class_from_uri(uri)

            if trait_class and trait_class.get_short_name() == name:
                if value is None:
                    self._properties.pop(uri, None)
                else:
                    self._properties[uri] = VitalSignsImpl.create_property_with_trait(prop_class, uri, value)
                return

        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")

    def my_getattr(self, name):
        if name == 'URI':
            return self._properties['http://vital.ai/ontology/vital-core#URIProp']
        for prop_info in self.get_allowed_properties():
            uri = prop_info['uri']
            # print(uri)
            trait_class = VitalSignsImpl.get_trait_class_from_uri(uri)
            # print(trait_class)
            if trait_class and trait_class.get_short_name() == name:
                if uri in self._properties:
                    return self._properties[uri]
                else:
                    return None
        return NotImplemented

    def get_property_value(self, property_uri):
        if property_uri == 'http://vital.ai/ontology/vital-core#URIProp':
            return self._properties['http://vital.ai/ontology/vital-core#URIProp']
        trait_class = VitalSignsImpl.get_trait_class_from_uri(property_uri)
        if not trait_class:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute with uri'{property_uri}'")
        name = trait_class.get_short_name()
        return self.__getattr__(name)

    def __getattr__(self, name):
        value = self.my_getattr(name)
        if value is NotImplemented:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")
        return value

    @classmethod
    def is_top_level_class(cls):
        return cls.__bases__ == (object,)

    @classmethod
    @abstractmethod
    def get_class_uri(cls) -> str:
        pass

    def to_json(self) -> str:

        serializable_dict = {}

        for uri, prop in self._properties.items():
            prop_value = prop.to_json()["value"]
            if uri == 'http://vital.ai/ontology/vital-core#URIProp':
                serializable_dict['URI'] = prop_value
            else:
                serializable_dict[uri] = prop_value

        class_uri = self.get_class_uri()

        serializable_dict['type'] = class_uri

        serializable_dict['types'] = [class_uri]

        return json.dumps(serializable_dict, indent=2)

    def to_rdf(self, format='nt', graph_uri: str = None) -> str:

        g = Graph(identifier=URIRef(graph_uri) if graph_uri else None)
        subject = URIRef(str(self._properties['http://vital.ai/ontology/vital-core#URIProp']))

        for prop_uri, prop_instance in self._properties.items():
            rdf_data = prop_instance.to_rdf()
            g.add((subject, URIRef(prop_uri), Literal(rdf_data["value"], datatype=rdf_data["datatype"])))

        return g.serialize(format=format)





import datetime
from vital_ai_vitalsigns_core.model.VitalServiceConfig import VitalServiceConfig


class VitalServiceAllegrographConfig(VitalServiceConfig):
        catalogName: str
        password: str
        poolMaxTotal: int
        repositoryName: str
        serverURL: str
        username: str




from vital_ai_vitalsigns.impl.vitalsigns_registry import VitalSignsRegistry


class VitalSignsMeta(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class VitalSigns(metaclass=VitalSignsMeta):
    def __init__(self):
        self._registry = VitalSignsRegistry()
        self._embedding_model_registry = {}
        self._registry.build_registry()

    def get_registry(self):
        return self._registry

    def put_embedding_model(self, name, model):
        """Add a model instance to the registry."""
        self._embedding_model_registry[name] = model

    def get_embedding_model(self, name):
        """Retrieve a model instance from the registry by its name."""
        return self._embedding_model_registry.get(name)







from rdflib import Graph, URIRef


class RdfCollectionImpl:
    def __init__(self):
        self.graph = Graph()

    def add_triples(self, nt_string: str):
        """
        Add multiple triples to the graph from an NT string.
        :param nt_string: Multiple triples in NT format.
        """
        self.graph.parse(data=nt_string, format="nt")

    def clear_graph(self):
        """
        Clear all triples from the graph.
        """
        self.graph = Graph()  # Reinitialize to clear

    def query_graph(self, sparql_query: str):
        """
        Query the graph using a SPARQL query.
        :param sparql_query: The SPARQL query string.
        :return: The query results.
        """
        return self.graph.query(sparql_query)

    def delete_triples(self, subject_uri: str):
        """
        Delete all triples with the specified subject URI.
        :param subject_uri: The subject URI as a string.
        """
        subject = URIRef(subject_uri)
        self.graph.remove((subject, None, None))



from vectordb import InMemoryExactNNVectorDB, HNSWVectorDB
from docarray import BaseDoc, DocList
from docarray.typing import NdArray
from typing import List, Union, TypeVar, Generic
from tempfile import TemporaryDirectory

T = TypeVar('T', bound=BaseDoc)


class VectorCollectionImpl:

    def __init__(self, schema: T):
        """
                Initializes the VectorCollectionImpl with a specific schema and workspace.
                :param schema: A subclass of BaseDoc defining the schema for the database.
                """
        self.schema = schema

        temp_dir = TemporaryDirectory()
        self.temp_dir = temp_dir

        # self.db = HNSWVectorDB[schema](workspace=temp_dir.name)
        self.db = InMemoryExactNNVectorDB[schema](workspace=temp_dir.name)

    def index_documents(self, documents: DocList[T]):
        """
        Indexes a list of documents in the vector database.
        :param documents: A DocList containing documents to be indexed.
        """
        self.db.index(documents)

    def search(self, query_embedding: NdArray, limit: int = 10) -> List[T]:
        """
        Searches the vector database for the closest matches to the provided embedding.
        :param query_embedding: A vector used to search for similar documents.
        :param limit: The maximum number of results to return.
        :return: A list of objects matching the query.
        """
        query_doc = self.schema(embedding=query_embedding)
        results = self.db.search(query_doc, limit=limit)
        return results if results else []

    def remove_doc(self, doc_id: str):
        self.db.delete([doc_id])



class PropertyConstraint:
    def __init__(self, expression):
        self.expression = expression

    def __repr__(self):
        return self.expression

    def __eq__(self, other):
        return PropertyConstraint(f"{self.expression} == {other}")

    def __ne__(self, other):
        return PropertyConstraint(f"{self.expression} != {other}")




from typing import Any, List

from vital_ai_vitalsigns.query.result_element import ResultElement


class ResultList(List[ResultElement]):
    def __init__(self, *args):
        super().__init__(*args)

    def add_result(self, graph_object: Any, score: float = 1.0):
        """
        Adds a ResultElement to the list.
        :param graph_object: The associated object (e.g., a document or data structure).
        :param score: The similarity score.
        """
        self.append(ResultElement(graph_object, score))



from vital_ai_vitalsigns.embedding.embedding_model import EmbeddingModel


def main():
    print('Hello World')

    embedder = EmbeddingModel()

    text1 = "This is a test sentence."
    vector1 = embedder.vectorize(text1)
    print(f"Vector for single sentence: {vector1}")

    texts = ["This is the first test sentence.", "Here is another one."]
    vectors = embedder.vectorize(texts)
    print(f"Vectors for multiple sentences: {vectors}")


if __name__ == "__main__":
    main()





from vital_ai_vitalsigns.model.GraphObject import GraphObject
from vital_ai_vitalsigns.model.properties.BooleanProperty import BooleanProperty
from vital_ai_vitalsigns.model.properties.LongProperty import LongProperty
from vital_ai_vitalsigns.model.properties.StringProperty import StringProperty
from vital_ai_vitalsigns.model.properties.URIProperty import URIProperty


class VITAL_HyperNode(GraphObject):

    _allowed_properties = [
        # {'uri': 'http://vital.ai/ontology/vital-core#URIProp', 'prop_class': URIProperty},
        {'uri': 'http://vital.ai/ontology/vital-core#isActive', 'prop_class': BooleanProperty},
        {'uri': 'http://vital.ai/ontology/vital-core#hasVersionIRI', 'prop_class': URIProperty},
        {'uri': 'http://vital.ai/ontology/vital-core#hasUpdateTime', 'prop_class': LongProperty},
        {'uri': 'http://vital.ai/ontology/vital-core#hasTimestamp', 'prop_class': LongProperty},
        {'uri': 'http://vital.ai/ontology/vital-core#hasProvenance', 'prop_class': URIProperty},
        {'uri': 'http://vital.ai/ontology/vital-core#hasOntologyIRI', 'prop_class': URIProperty},
        {'uri': 'http://vital.ai/ontology/vital-core#hasName', 'prop_class': StringProperty},
        {'uri': 'http://vital.ai/ontology/vital-core#vitaltype', 'prop_class': URIProperty},
        {'uri': 'http://vital.ai/ontology/vital-core#types', 'prop_class': URIProperty}
    ]

    @classmethod
    def get_allowed_properties(cls):
        return super().get_allowed_properties() + VITAL_HyperNode._allowed_properties

    @classmethod
    def get_class_uri(cls) -> str:
        return 'http://vital.ai/ontology/vital-core#VITAL_HyperNode'



from typing import Any, List


class ResultElement:
    def __init__(self, graph_object: Any, score: float = 1.0):
        self.graph_object = graph_object
        self.score = score

    def __repr__(self):
        return f"ResultElement(graph_object={self.graph_object},score={self.score})"




import datetime
from vital_ai_vitalsigns.model.VITAL_Node import VITAL_Node


class RDFStatement(VITAL_Node):
        rdfContext: str
        rdfObject: str
        rdfPredicate: str
        rdfSubject: str





from vital_ai_vitalsigns.model.properties.BooleanProperty import BooleanProperty
from vital_ai_vitalsigns.model.properties.DateTimeProperty import DateTimeProperty
from vital_ai_vitalsigns.model.properties.DoubleProperty import DoubleProperty
from vital_ai_vitalsigns.model.properties.FloatProperty import FloatProperty
from vital_ai_vitalsigns.model.properties.GeoLocationProperty import GeoLocationProperty
from vital_ai_vitalsigns.model.properties.IntegerProperty import IntegerProperty
from vital_ai_vitalsigns.model.properties.LongProperty import LongProperty
from vital_ai_vitalsigns.model.properties.OtherProperty import OtherProperty
from vital_ai_vitalsigns.model.properties.StringProperty import StringProperty
from vital_ai_vitalsigns.model.properties.TruthProperty import TruthProperty
from vital_ai_vitalsigns.model.properties.URIProperty import URIProperty
from vital_ai_vitalsigns_core.model.VitalServiceConfig import VitalServiceConfig


class VitalServiceIndexedDBConfig(VitalServiceConfig):
    _allowed_properties = [
        {'uri': 'http://vital.ai/ontology/vital-core#hasGraphQueries', 'prop_class': StringProperty}, 
        {'uri': 'http://vital.ai/ontology/vital-core#hasSelectQueries', 'prop_class': StringProperty}, 
    ]

    @classmethod
    def get_allowed_properties(cls):
        return super().get_allowed_properties() + VitalServiceIndexedDBConfig._allowed_properties

    @classmethod
    def get_class_uri(self) -> str:
        return 'http://vital.ai/ontology/vital-core#VitalServiceIndexedDBConfig'




from vital_ai_vitalsigns.query.vital_query import VitalQuery


class VitalGraphQuery(VitalQuery):
    def __init__(self):
        self.segments = []
        self.arcs = []

    def add_segment(self, segment):
        self.segments.append(segment)
        return self

    def add_arc(self, arc):
        self.arcs.append(arc)
        return self

    def __repr__(self):
        return f"GraphQuery(segments={self.segments}, arcs={self.arcs})"


class Assignment:
    def __init__(self, variable):
        self.variable = variable

    def to(self, value):
        if isinstance(value, type):
            value = value.__name__
        return f"{self.variable} = {value}"


def bind(variable):
    return Assignment(variable)


def ref(variable):
    return f"?{variable}"


class Arc:
    def __init__(self):
        self.node_constraints = []
        self.edge_constraints = []
        self.constraints = []
        self.provisions = []
        self.arcs = []

    def node_constraint(self, constraint):
        if isinstance(constraint, type):
            constraint = f"{constraint.__name__}"
        self.node_constraints.append(constraint)
        return self

    def edge_constraint(self, constraint):
        if isinstance(constraint, type):
            constraint = constraint().__repr__()
        self.edge_constraints.append(constraint)
        return self

    def constraint(self, constraint):
        self.constraints.append(constraint)
        return self

    def provides(self, provision):
        self.provisions.append(provision)
        return self

    def add_arc(self, arc):
        self.arcs.append(arc)
        return self

    def __repr__(self):
        return (f"Arc(node_constraints={self.node_constraints}, "
                f"edge_constraints={self.edge_constraints}, "
                f"constraints={self.constraints}, "
                f"provisions={self.provisions}, arcs={self.arcs})")

    def to_sparql(self):
        sparql = ""
        return sparql


class ArcGroup:
    pass


class ArcOr(ArcGroup):
    def __init__(self):
        self.arcs = []

    def add_arc(self, arc):
        self.arcs.append(arc)
        return self

    def __repr__(self):
        return f"ArcOr(arcs={self.arcs})"


class ArcAnd(ArcGroup):
    def __init__(self):
        self.arcs = []

    def add_arc(self, arc):
        self.arcs.append(arc)
        return self

    def __repr__(self):
        return f"ArcAnd(arcs={self.arcs})"




from vital_ai_vitalsigns.model.VITAL_Edge import VITAL_Edge
from vital_ai_vitalsigns.model.VITAL_Node import VITAL_Node
from vital_ai_vitalsigns.query.vital_graph_query import VitalGraphQuery, Arc, ref, ArcAnd, bind
from vital_ai_vitalsigns.query.vital_select_query import VitalSelectQuery
from vital_ai_vitalsigns_core.model.properties.Property_URIProp import Property_URIProp


def main():
    print('Hello World')

    # MetaQL Select Query
    """
    SELECT {
        value limit: 100
        value offset: 0
        value segments: ["mydata"]
        constraint { Person.class }
        constraint { Person.props().name.equalTo("John" ) }
    }
    """

    select_query = (VitalSelectQuery()
                    .add_segment("mydata")
                    .node_constraint(VITAL_Node)
                    .edge_constraint(VITAL_Edge)
                    .node_constraint(VITAL_Node.name == "John")
                    )

    print(select_query)

    # MetaQL: Graph Query
    """
    GRAPH {
        value segments: ["mydata"]
        ARC {
          node_constraint { Email.class }
          constraint { "?person1 != ?person2" }
          ARC_AND {
              ARC {
                edge_constraint { Edge_hasSender.class }
                node_constraint { Person.props().emailAddress.equalTo("john@example.org")
                node_constraint { Person.class }
                node_provides { "person1 = URI" }
             }
              ARC {
                edge_constraint { Edge_hasRecipient.class }
                node_constraint { Person.class }
                node_provides { "person2 = URI" }
             }
          }
        }
    }
    """

    graph_query = (VitalGraphQuery()
             .add_segment("mydata")
             .add_arc(Arc()
                      .node_constraint(VITAL_Node)
                      .constraint(ref("person1") != ref("person2"))
                      .add_arc(ArcAnd()
                               .add_arc(Arc()
                                        .edge_constraint(VITAL_Edge)
                                        .node_constraint(VITAL_Node.name == "John")
                                        .node_constraint(VITAL_Node)
                                        .provides(bind("person1").to(Property_URIProp)))
                               .add_arc(Arc()
                                        .edge_constraint(VITAL_Edge)
                                        .node_constraint(VITAL_Node)
                                        .provides(bind("person2").to(Property_URIProp)))
                               )
                      )
             )

    print(graph_query)

    # sparql = graph.to_sparql()
    # print(sparql)


if __name__ == "__main__":
    main()




class VitalQuery:
    pass


class Constraint:
    def __init__(self, expression):
        self.expression = expression

    def __repr__(self):
        return self.expression

    def __eq__(self, other):
        return Constraint(f"{self.expression} == {other}")

    def __ne__(self, other):
        return Constraint(f"{self.expression} != {other}")


class Field:
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return f"?{self.name}"

    def __getattr__(self, item):
        return Constraint(f"?{self.name}.{item}")



from vital_ai_vitalsigns.query.vital_query import VitalQuery


class VitalSelectQuery(VitalQuery):
    def __init__(self):
        self.segments = []
        self.node_constraints = []
        self.edge_constraints = []

    def add_segment(self, segment):
        self.segments.append(segment)
        return self

    def __repr__(self):
        return f"SelectQuery(segments={self.segments}, node_constraints={self.node_constraints}, edge_constraints={self.edge_constraints})"

    def node_constraint(self, constraint):
        if isinstance(constraint, type):
            constraint = f"{constraint.__name__}"
        self.node_constraints.append(constraint)
        return self

    def edge_constraint(self, constraint):
        if isinstance(constraint, type):
            constraint = constraint().__repr__()
        self.edge_constraints.append(constraint)
        return self





import logging
from vital_ai_vitalsigns.collection.graph_collection import GraphCollection
from vital_ai_vitalsigns.embedding.embedding_model import EmbeddingModel
from vital_ai_vitalsigns.model.VITAL_Node import VITAL_Node
from vital_ai_vitalsigns.vitalsigns import VitalSigns
from vital_ai_vitalsigns_core.model.properties.Property_hasName import Property_hasName


def main():
    logging.basicConfig(level=logging.INFO)

    logging.info('Hello World')

    embedder = EmbeddingModel()

    vs = VitalSigns()
    vs.put_embedding_model(embedder.get_model_id(), embedder)

    graph = GraphCollection()

    class_uri = VITAL_Node().get_class_uri()

    property_uris = [Property_hasName.get_uri()]

    graph.set_vector_properties(class_uri, property_uris)

    node1 = VITAL_Node()
    node1.URI = 'uri:node1'
    node1.name = 'Omelet'

    node2 = VITAL_Node()
    node2.URI = 'uri:node2'
    node2.name = 'Pork Chop'

    node3 = VITAL_Node()
    node3.URI = 'uri:node3'
    node3.name = 'Coffee'

    node4 = VITAL_Node()
    node4.URI = 'uri:node4'
    node4.name = 'Orange Juice'

    node5 = VITAL_Node()
    node5.URI = 'uri:node5'
    node5.name = 'Spaghetti'

    graph.add_objects([node1, node2, node3, node4, node5])

    node = graph.get('uri:node2')

    print('Look up Node: ' + str(node.name))

    # query = 'pig'
    # query = 'apple'
    # query = 'noodles'
    # query = 'tea'
    query = 'frittata'

    logging.info('Search for: ' + query)

    results = graph.search(query)

    if results is not None:
        re = results[0]
        obj = re.graph_object
        score = re.score
        logging.info('Top Result for: ' + query + ': ' + str(obj.name) + " (" + str(score) + ")")
        logging.info(obj.to_json())
        logging.info(obj.to_rdf())
    else:
        logging.info('No results')

    # for r in results:
    #    logging.info(r)

    sparql_results = graph.sparql_query("SELECT ?s ?p ?o WHERE { ?s ?p ?o } ORDER BY ?s ?p")

    sparql_result_count = len(sparql_results)
    logging.info('Sparql Result Count: ' + str(sparql_result_count))

    for row in sparql_results:
        logging.info(f"Subject: {row['s']}, Property: {row['p']}, Object: {row['o']}")


if __name__ == "__main__":
    main()



from .IProperty import IProperty


class TruthProperty(IProperty):
    def __init__(self, value: str):
        super().__init__(value)

    def __str__(self):
        return str(self.value)

    def __eq__(self, other):
        if isinstance(other, str):
            return self.value == other
        elif isinstance(other, TruthProperty):
            return self.value == other.value
        return NotImplemented

    def __rshift__(self, other):
        return self == other

    def to_json(self):
        return {"value": self.value}



from setuptools import setup, find_packages

setup(
    name='vital-ai-vitalsigns',
    version='0.1.5',
    author='Marc Hadfield',
    author_email='marc@vital.ai',
    description='VitalSigns knowledge graph bindings',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/vital-ai/vital-vitalsigns-python',
    packages=find_packages(exclude=["test"]),
    entry_points={
        'vitalsigns_packages': [
            'vital_ai_vitalsigns = vital_ai_vitalsigns',
            'vital_ai_vitalsigns_core = vital_ai_vitalsigns_core'
        ]
    },
    package_data={
         '': ['*.pyi'],
    },
    license='Apache License 2.0',
    install_requires=[
        'sentence_transformers==2.7.0',
        'vectordb== 0.0.21',
        'docarray==0.40.0',
        'rdflib==7.0.0'
    ],
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10',
)



from vital_ai_vitalsigns.model.GraphObject import GraphObject


class VITAL_HyperNode(GraphObject):
    # URIProp: str
    URI: str
    active: bool
    versionIRI: str
    updateTime: int
    timestamp: int
    provenance: str
    ontologyIRI: str
    name: str
    vitaltype: str
    types: str




from sentence_transformers import SentenceTransformer
from typing import List, Union, Dict
import hashlib
from collections import OrderedDict


class LRUEmbeddingCache:
    def __init__(self, maxsize=1000):
        self.cache = OrderedDict()
        self.maxsize = maxsize

    def get(self, key: str):
        if key in self.cache:
            self.cache.move_to_end(key)
            return self.cache[key]
        return None

    def set(self, key: str, value: List[float]):
        if key in self.cache:
            self.cache.move_to_end(key)
        self.cache[key] = value
        if len(self.cache) > self.maxsize:
            self.cache.popitem(last=False)


class EmbeddingModel:
    def __init__(self, model_id: str = 'paraphrase-MiniLM-L3-v2', cache_size=1000):
        # self.model = SentenceTransformer('paraphrase-multilingual-MiniLM-L12-v2')
        # self.model = SentenceTransformer('all-MiniLM-L6-v2')
        self.model_id = model_id
        self.model = SentenceTransformer(model_id)
        self.cache = LRUEmbeddingCache(maxsize=cache_size)

    def get_model_id(self):
        return self.model_id

    def hash_text(self, text: str) -> str:
        # Create a SHA-256 hash of the text
        return hashlib.sha256(text.encode('utf-8')).hexdigest()

    def vectorize(self, text: Union[str, List[str]]) -> Union[List[float], List[List[float]]]:
        """
        Vectorize a single string or a list of strings using the model.
        :param text: A string or a list of strings to be vectorized.
        :return: A list of floats (if single string) or list of list of floats (if multiple strings).
        """
        single_string = False
        if isinstance(text, str):
            text = [text]
            single_string = True

        # Create a list to hold results
        results = []
        uncached_texts = []

        for t in text:
            hash_key = self.hash_text(t)
            cached_vector = self.cache.get(hash_key)
            if cached_vector is not None:
                results.append(cached_vector)
            else:
                uncached_texts.append(t)

        if uncached_texts:
            new_embeddings = self.model.encode(uncached_texts)
            for t, vector in zip(uncached_texts, new_embeddings):
                hash_key = self.hash_text(t)
                self.cache.set(hash_key, vector)
                results.append(vector)

        if single_string:
            return results[0]

        return results



from importlib.metadata import entry_points
import importlib
import pkgutil
import logging

# these are not imported here to remove a circular dependency on start-up
# from vital_ai_vitalsigns.model.VITAL_Edge import VITAL_Edge
# from vital_ai_vitalsigns.model.VITAL_GraphContainerObject import VITAL_GraphContainerObject
# from vital_ai_vitalsigns.model.VITAL_HyperEdge import VITAL_HyperEdge
# from vital_ai_vitalsigns.model.VITAL_HyperNode import VITAL_HyperNode
# from vital_ai_vitalsigns.model.VITAL_Node import VITAL_Node
# from vital_ai_vitalsigns.model.trait.PropertyTrait import PropertyTrait


class VitalSignsRegistry:

    def __init__(self):
        self.vitalsigns_packages = []
        self.vitalsigns_classes = {}
        self.vitalsigns_property_classes = {}
        self.vitalsigns_ontologies = []

    def build_registry(self):
        self.vitalsigns_packages = []
        self.vitalsigns_ontologies = []

        self.vitalsigns_classes = {}

        logging.info('building vitalsigns class and property registry...')

        from vital_ai_vitalsigns.model.VITAL_Edge import VITAL_Edge
        from vital_ai_vitalsigns.model.VITAL_GraphContainerObject import VITAL_GraphContainerObject
        from vital_ai_vitalsigns.model.VITAL_HyperEdge import VITAL_HyperEdge
        from vital_ai_vitalsigns.model.VITAL_HyperNode import VITAL_HyperNode
        from vital_ai_vitalsigns.model.VITAL_Node import VITAL_Node

        self.vitalsigns_classes[VITAL_Node.get_class_uri()] = VITAL_Node
        self.vitalsigns_classes[VITAL_Edge.get_class_uri()] = VITAL_Edge
        self.vitalsigns_classes[VITAL_HyperNode.get_class_uri()] = VITAL_HyperNode
        self.vitalsigns_classes[VITAL_HyperEdge.get_class_uri()] = VITAL_HyperEdge
        self.vitalsigns_classes[VITAL_GraphContainerObject.get_class_uri()] = VITAL_GraphContainerObject

        self.vitalsigns_property_classes = {}

        for ep in entry_points().get('vitalsigns_packages', []):
            try:
                module = importlib.import_module(ep.value)
                self.vitalsigns_packages.append(module)
                logging.info('Discovered Module: ' + str(module))
            except ImportError as e:
                logging.info(f"Could not import {ep.value}: {e}")
        for p in self.vitalsigns_packages:
            self.scan_vitalsigns_classes(p)

        logging.info('completed build of vitalsigns class and property registry.')

    def _scan_module(self, module_name):
        try:
            module = importlib.import_module(module_name)
        except ImportError as e:
            logging.info(f"Error importing {module_name}: {e}")
            return

        for cls_name, cls in vars(module).items():
            if isinstance(cls, type):
                if self.is_vitalsigns_ontology_class(cls):
                    logging.info(f"Found VitalSigns Domain Ontology Class {cls_name} URI: {cls.OntologyURI}")
                    self.vitalsigns_ontologies.append(cls)
                if self.is_vitalsigns_class(cls):
                    class_uri = cls.get_class_uri()
                    logging.info(f"Found VitalSigns Graph Class {cls_name} URI: {class_uri}")
                    self.vitalsigns_classes[class_uri] = cls
                if self.is_vitalsigns_property_class(cls):
                    property_trait_uri = cls.get_uri()
                    logging.info(f"Found VitalSigns Property Class {cls_name} URI: {property_trait_uri}")
                    self.vitalsigns_property_classes[property_trait_uri] = cls

    def scan_vitalsigns_classes(self, package):
        for finder, name, is_pkg in pkgutil.walk_packages(package.__path__, package.__name__ + '.'):
            self._scan_module(name)
            if is_pkg:
                self._scan_module(name)

    def is_vitalsigns_ontology_class(self, cls):

        from vital_ai_vitalsigns.model.BaseDomainOntology import BaseDomainOntology

        if cls is BaseDomainOntology:
            return False
        if issubclass(cls, BaseDomainOntology):
            return True

    def is_vitalsigns_class(self, cls):

        from vital_ai_vitalsigns.model.VITAL_Edge import VITAL_Edge
        from vital_ai_vitalsigns.model.VITAL_GraphContainerObject import VITAL_GraphContainerObject
        from vital_ai_vitalsigns.model.VITAL_HyperEdge import VITAL_HyperEdge
        from vital_ai_vitalsigns.model.VITAL_HyperNode import VITAL_HyperNode
        from vital_ai_vitalsigns.model.VITAL_Node import VITAL_Node

        if cls is VITAL_Node:
            return False
        if cls is VITAL_Edge:
            return False
        if cls is VITAL_HyperNode:
            return False
        if cls is VITAL_HyperEdge:
            return False
        if cls is VITAL_GraphContainerObject:
            return False
        if issubclass(cls, VITAL_Node):
            return True
        if issubclass(cls, VITAL_Edge):
            return True
        if issubclass(cls, VITAL_HyperNode):
            return True
        if issubclass(cls, VITAL_HyperEdge):
            return True
        if issubclass(cls, VITAL_GraphContainerObject):
            return True
        return False

    def is_vitalsigns_property_class(self, cls):

        from vital_ai_vitalsigns.model.trait.PropertyTrait import PropertyTrait

        if cls is PropertyTrait:
            return False
        if issubclass(cls, PropertyTrait):
            return True
        return False



class VitalSignsPackage:
    vitalsigns_core_package = 'vital_ai_vitalsigns_core'
    vital_domain_package = 'vital_ai_domain'

    # TODO add lookup methods for registered domains
    @classmethod
    def get_package(cls, owl_url: str):
        return 'ai_vital_aimp'




import datetime
from vital_ai_vitalsigns_core.model.DatabaseConnection import DatabaseConnection


class SparqlDatabaseConnection(DatabaseConnection):
        catalogName: str
        repositoryName: str

